%
% Author : Ivor W. Tsang
%
% PreImaging
%
% This code is free for research purpose, and must not be further distributed 
% without prior permission of the authors. If you use it in your scientific 
% work, please cite as 
%
% James T. Kwok, Ivor W. Tsang. The pre-image problem in kernel methods. IEEE 
% Transactions on Neural Networks, 15(6):1517-1525, Nov 2004.
%

function resX=PreImaging(d2,X,nb_size)

% Input Argument
% d2      : a Nx1 vector of distances between x and x_i
% X       : a dxN matrix of N points x_i in d dimension
% nb_size : the size of neighbors
%
% Output Argument
% resX    : the resultant preimage of x


% find out the closet neighbourlood
[sd2,sidx]=sort(d2);
    
% centering data in input space
XH    = X(:,sidx(1:nb_size));
Xmean = mean(XH,2);               % the center of neighbors
XH    = XH-Xmean*ones(1,nb_size);

% SVD decompose
[UM,WM,VM] = svd(XH,0);
rankM      = rank(WM,1e-5);
UM         = UM(:,1:rankM);
WM         = WM(1:rankM,1:rankM);
VM         = VM(:,1:rankM);
transf     = UM*(diag(1./diag(WM))*(VM'./2));

% denoising
sd0 = zeros(nb_size,1);
ZM  = WM*VM';
for i=1:nb_size
    sd0(i) = ZM(:,i)'*ZM(:,i);
end

resX = transf*(sd0-sd2(1:nb_size))+Xmean;               
