%
% Author : Ivor W. Tsang, Jooyoung Park
%
% kpca_denoising_MDS
%
% This code is free for research purpose, and must not be further distributed 
% without prior permission of the authors. If you use it in your scientific 
% work, please cite as 
%
% James T. Kwok, Ivor W. Tsang. The pre-image problem in kernel methods. IEEE 
% Transactions on Neural Networks, 15(6):1517-1525, Nov 2004.
%
% Jooyoung Park, Daesung Kang, Jongho Kim, James T. Kwok, Ivor W. Tsang. 
% SVDD-Based Pattern De-Noising. Neural Computation, 19(7): 1919-1938, July 
% 2007
%

function testX_denoised = kpca_denoising_MDS(X,testX,nb_size,max_ev,rbf_var,iter_num)

% Input Arguments
% X        : "d x num_of_train" matrix of (prototype) training data
% testX    : "d x num_of_test" matrix of (corrupted) test data
% max_ev   : number of principal components
% rbf_var  : kernel parameter of rbf kernel
% iter_num : iteration number for de-noising
%
% Output Argument
% testX_denoised : the denoised x


[dim,num_trn]  = size(X);
[dim,num_test] = size(testX);

% H is centering matrix
H = eye(num_trn)-ones(num_trn,num_trn)/num_trn;    

% compute the kernel matrix
K = zeros(num_trn,num_trn);
for i_th=1:num_trn,
    for j_th=i_th:num_trn,   
        K(i_th,j_th) = exp(-sum((X(:,i_th)-X(:,j_th)).^2)/(rbf_var));
        K(j_th,i_th) = K(i_th,j_th);
    end
end
K_n   = H*K*H;

% perform kernel pca
[evecs,evals,percents] = pcacov(K_n);

%%%%%% eigenvalue %%%%%%%%%%%%%%%%%%%%%
max_ev;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
evecs = evecs(:,1:max_ev);
evals = evals(1:max_ev);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% normalize the eigenvectors in feature space 
% alpha is a normalized eigenvector matrix
alpha = zeros(num_trn, max_ev);
for k = 1:max_ev
    alpha(:,k) = evecs(:,k)/sqrt(evals(k));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

HMH   = H*alpha*alpha'*H;        
cK    = mean(K,2);
Kmean = mean(cK);

testX_denoised = zeros(size(testX));
testpt         = testX;
for (iter=1:iter_num),
    for j_th = 1:num_test    
    % calculate the distance between the testing point and training points
        k_x       = test_kernel(X,testpt(:,j_th),rbf_var);
        gammaC    = HMH*(k_x-cK);
        PphiNormC = (k_x+cK)'*gammaC+Kmean;
        d2        = zeros(num_trn,1);
        for i_th = 1 : num_trn
            PphiProjC = K(i_th,:)*gammaC+cK(i_th);
            d2(i_th)  = -log((1-PphiNormC+2*PphiProjC)/2)*rbf_var;
        end           
        testpt(:,j_th) = PreImaging(d2,X,nb_size);
    end
end
testX_denoised = testpt;
    